
#ifndef NANOCLIENT_H_
#define NANOCLIENT_H_

#include "nanolink.h"

#include <string>
using namespace std;

/*
功能:
	封装 Nanolink SDK 基础接口 并 增加事件接口

用法:
	#1, 从 NanoClient 派生 MyClient, 实现必要的 onMessage 事件接口
	class MyClient : public NanoClient {
		// 收到数据
		virtual void onMessage(uint8_t *data, int size, uint8_t fromIndex) {
			...
		}

		// 收到resync指令, 发送完整数据
		// **多人连接需要实现该接口**
		virtual void onResync (uint8_t fromIndex) {
			...
		}
	};

	#2, 实例化 MyClient 与 初始化
	MyClient client; //

	client.init("***appKey***", 2);
	client.config("time-machine", "true");

	#3, 连接 (通常在"开始"按钮触发)
	client.connect(1);

	#4, 在 Update 中调用 doUpdate 接口接收数据处理状态
	client.doUpdate ();

	#5, 发送数据 (在事件产生的地方序列化数据并调用 send)
	client.send (byte* data, int len);

	#6, 断开连接 (通常在战斗结束的时候)
	client.disconnect ();

	#7, 释放内存
	client.term ();
*/
class NanoClient {
protected:
	static string status; // 上次的状态

	static bool disconnectedBySelf; // 该客户端自己断开

	static int players; // 当前玩家数量
	static int last_times [64]; // 玩家上次事件时间 (多少 ms 之前), 用于判断 离开, 回来

private:
	// XXX 优化: 避免 getString, recv 频繁分配内存
	static int _recv_buf_size;
	static uint8_t* _recv_buf;

	static int _get_string_buf_size;
	static char* _get_string_buf;

private:
	// 功能:
	//	发送一个协议"分隔符"标记, 用在多人联网时
	//	** 已经在 onResync 前自动调用, 不需要再调用该接口
	//
	// 参数:
	//	mark: 标记值, 保留, 通常为0
	//
	// 返回值:
	// 	成功: >= 0, 发送数据的索引值
	//	失败: -1, 无法添加任务到发送队列满
	//		可以用 config ("send-tasks-size", "xxx") 调整发送队列大小
	static int _sendmark (uint8_t mark = 0) {
		return nano_sendmark (mark);
	}

public:
	// 功能:
	//	SDK 初始化
	//
	// 参数:
	//	appKey: 应用ID
	//	mode: SDK 模式
	//
	//	mode 定义:
	// 		0: 回放
	// 		1: 单人, 仅用于记录
	// 		2: 1对1联网方式(缺省)
	// 		3: 多人联网方式
	//
	// 返回值:
	//	成功: 0
	//	失败: -1, mode 参数错误时失败
	static int init (const char* appKey, int mode = 2) {
		disconnectedBySelf = false;

		players = 0;
		for (int i = 0; i < 64; i ++)
			last_times [i] = -1;

		return nano_init (appKey, mode);
	}

	// 功能:
	//	释放内存
	static int term () {
		if (_recv_buf)
			delete [] _recv_buf;
		if (_get_string_buf)
			delete [] _get_string_buf;
		return 0; // nano_term ();
	}

	// 功能:
	//	设置 协议中的一些参数, 具体参看 "Nanolink SDK 参数设置" 文档
	//
	// 参数:
	//	name: 参数名称
	//	value: 参数值
	//
	// 返回值:
	//	成功: 0
	// 	失败: -1, 不支持的参数名称
	static int config (const char* name, const char* value) {
		return nano_config (name, value);
	}

	// 功能:
	//	基于等级匹配连接(外网)
	//
	// 参数:
	// 	level: 当前客户端的等级, 具体匹配等级范围在服务器设置
	//	mode: 等级匹配可以分为不同模式, 支持 0-15 模式, 缺省为0
	//		例如, mode=0, 代表 经典模式, mode=1, 代表冒险模式, ...
	// 	timeout: 匹配等待时间, 单位秒，默认15秒
	//
	// 返回值:
	//	成功: 0
	//	失败: -1, init 接口没有调用 或 端口打开失败
	static int connect (uint8_t level, uint8_t mode = 0, int32_t timeout=15) {
		return nano_connect (level, mode, timeout*1000);
	}

	// 功能:
	//	基于名称匹配连接
	//
	// 参数:
	// 	group: 匹配的房间名称, 如果 group 为空，局域网匹配
	// 	timeout: 匹配等待时间, 单位秒，默认15秒
	//
	// 返回值:
	//	成功: 0
	//	失败: -1, init 接口没有调用 或 端口打开失败
	static int connect (const char* group = NULL, int32_t timeout=15) {
		return nano_connect2 (group, timeout*1000);
	}

	// 功能:
	//	断开当前连接
	//
	// 返回值:
	//	成功: 0
	// 	失败: -1, init 接口没有调用
	static int disconnect () {
		disconnectedBySelf = true;

		return nano_disconnect ();
	}

	// 功能:
	//	发送数据
	//
	// 参数:
	//	data: 待发送的数据
	//	size: 发送数据的长度
	//
	// 返回值:
	// 	成功: >= 0, 发送数据的索引值
	//	失败: -1, 无法添加任务到发送队列满
	//		可以用 config ("send-tasks-size", "xxx") 调整发送队列大小
	static int  send (uint8_t* data, uint32_t size, uint8_t strength=255) {
		return nano_send (data, size, strength);
	}

	// 功能:
	//	接收数据
	//	建议用 doUpdate + onMessage 方式接收处理数据
	//
	// 参数:
	//	data: 接收数据缓冲区
	//	size: 接收数据缓冲区大小
	// 	clientIndex: 发送者的索引值
	//
	// 返回值:
	// 	返回数据的长度
	//	-1: 没有收到新的数据
	//	0: 收到 resync 指令
	//	1: 收到 mark 指令
	//	>1: 收到用户指令
	static int recv (uint8_t* data, uint32_t size, uint8_t& clientIndex) {
		return nano_recv (data, size, clientIndex);
	}

	// 功能:
	//	更新客户端的位置(x,y), 空间(z), 接收增益(k)
	//
	// 参数:
	//	x, y 玩家的位置
	//	z 玩家所在空间, 缺省为0, 0-255, 不同空间互相不会收到数据
	//	k 接收增益, 接收范围的放大倍数
	//
	// 返回值:
	//	成功返回 0
	static int update (uint8_t x, uint8_t y, uint8_t z=0, uint8_t k=1) {
		return nano_update (x, y, z, k);
	}

	// 功能:
	//	加载日志数据, 用于日志回放
	//
	// 参数:
	//	filePath: 存放日志数据的文件路径
	//
	// 返回值:
	//	成功: 返回文件长度
	//	失败: -1, init 没有调用 或 文件打开失败
	static int load (const char* filePath) {
		return nano_load ((uint8_t*)filePath, -1);
	}

	// 功能:
	//	保存日志数据
	//	需要先启用日志功能 config ("time-machine", "true");
	//
	// 参数:
	//	filePath: 存放日志数据的文件路径
	//
	// 返回值:
	//	成功: 返回文件长度
	//	失败: -1, init 没有调用 或 文件打开失败
	static int save (const char* filePath) {
		return nano_save ((uint8_t*)filePath, -1);
	}

	// 功能:
	//	获取协议中的一些数值(整数类型)
	//
	// 参数:
	// 	name: players, latency, last-time, last-error, client-index, is-master, master-index, status
	//  n: last-time 时配合使用, 缺省-1
	//
	// 返回值:
	//	该参数的数值, 如果该参数不存在, 返回0
	static int getInt (const char* name, int n = -1) {
		if (strcmp (name, "players") == 0) {
			if (n == -1)
				return players;
			else {
				// 房间内在线玩家数量, "在线" 根据 n(timeout) 判断
				int count = 0;
				for (int i = 0; i < players; i ++) {
					int v = getInt ("last-time", i);
					if (v >= 0 && v < n) {
						count ++;
					}
				}

				return count;
			}
		}

		return nano_getInt (name, n);
	}

	// 功能:
	//	获取协议中的一些数值(字符串类型)
	//
	// 参数:
	// 	name: server-id, target-id, client-id, status, stats
	//
	// 返回值:
	//	该参数的数值, 如果该参数不存在, 返回""
	static string getString (const char* name) {
		int n = nano_getString (name, _get_string_buf, _get_string_buf_size);
		if (n <= 0) {
			return string ("");
		}

		if (n > _get_string_buf_size) {
			_get_string_buf_size *= 2;
			if (_get_string_buf_size < n)
				_get_string_buf_size = n;

			if (_get_string_buf)
				delete [] _get_string_buf;

			_get_string_buf = new char [_get_string_buf_size];

			n = nano_getString (name, _get_string_buf, _get_string_buf_size);
		}

		return string (_get_string_buf, n);
	}

	// 功能:
	//	获得当前协议状态
	//
	// 返回值:
	//	init, matching, connecting, connected, disconnect
	static string getStatus () {
		return getString ("status");
	}

	// 功能:
	//	判断连接是否建立
	//
	// 返回值:
	//	true: 已经建立连接
	//	false: 还没有建立连接
	static bool isConnected() {
		return (getStatus () == "connected");
	}

	// 功能:
	//	判断连接是否断开
	//
	// 返回值:
	//	true: 已经断开连接
	//	false: 还没有断开连接
	static bool isDisconnect() {
		return (getStatus () == "disconnect");
	}

public:
	// 功能:
	//	处理状态变化与数据接收, 分发到接口的各个 on 事件中
	//
	// 返回值:
	//	>=0: 接收并分发的数据条数
	//	-1: 连接还没有建立
	virtual int doUpdate () {
		// 状态变化时, 调用相应状态变化函数
		if (status != getStatus ()) {
			string newStatus = getStatus ();
			if (newStatus == "connected") {
				onConnected ();

				_sendmark (0);
				onResync (getInt ("client-index"));
			}
			if (newStatus == "disconnect") {
				onDisconnected (getInt("last-error"));
			}

			onStatusChanged (getStatus (), status);

			status = newStatus;
		}

		// 还没有连接成功，忽略后续的接收处理
		if (!isConnected ())
			return -1;

		// 检查是否有人 "离开" 或 "回来"
		int timeout = 2000;
		for (int i = 0; i < players; i ++) {
			int old_last_time = last_times [i];
			last_times [i] = getInt ("last-time", i);
			
			if (old_last_time < timeout &&  last_times [i] > timeout) {
				onPlayer (i, "left"); // 离开
			} else if (old_last_time > timeout &&  last_times [i] < timeout) {
				onPlayer (i, "return"); // 回来
			}
		}

		uint8_t fromIndex;
		
		int count = 0;
		while (true) {
			int len = nano_recv (_recv_buf, _recv_buf_size, fromIndex);

			// 没有收到指令或数据, 退出
			if (len < 0)
				break;

			// 如果空间不足, 重新分配
			if (len > _recv_buf_size) {
				_recv_buf_size *= 2;
				if (_recv_buf_size < len)
					_recv_buf_size = len;

				if (_recv_buf)
					delete [] _recv_buf;

				_recv_buf = new uint8_t [_recv_buf_size];

				len = nano_recv (_recv_buf, _recv_buf_size, fromIndex);
			}

			if (fromIndex >= players) {
				// 有可能 msg.fromIndex 一下子比 this.players 大不少
				for (int i = players; i < fromIndex+1; i ++) {
					last_times [i] = getInt ("last-time", i);

					onPlayer (i, "new");
				}
				players = fromIndex+1;
			}

			if (len == 0) { // 收到同步指令
				_sendmark (0);
				onResync ((uint8_t) fromIndex);
			} else if (len == 1) { // 收到标记指令
				// onMark (buf [0], fromIndex);
			} else { // 收到数据
				onMessage (_recv_buf, len, fromIndex);
			}

			count ++;
		}

		return count;
	}

	// 功能:
	//	接收到一个消息时的处理接口
	//
	// 参数:
	//	data: 接收数据缓冲区
	//	size: 数据大小
	//	fromIndex: 发送者
	//
	// 用法:
	//	收到的数据进行解码, 转换成不同指令, 并分发给对应的指令代码
	virtual void onMessage (uint8_t* data, int size, uint8_t fromIndex) {
	}

	// 功能:
	//	收到 resync 指令时的处理接口
	//
	// 参数:
	//	fromIndex: 发送者
	//	如果是自己(建立连接时触发), fromIndex = getInt ("client-index")
	//
	// 用法:
	//  向其他客户端重新发送当前客户端的完整数据
	//	包括 昵称, 颜色, ... 等只在初始化传输的数据
	virtual void onResync (uint8_t fromIndex) {
	}

	// 功能:
	//	状态变化
	//
	// 参数:
	//	newStatus: 变化后的状态
	//	oldStatus: 变化前的状态
	virtual void onStatusChanged (string newStatus, string oldStatus) {
	}

	// 功能:
	//	连接建立
	//
	// 用法:
	//	连接建立后初始化, 开始游戏
	virtual void onConnected () {
	}

	// 功能:
	//	连接断开
	//
	// 参数:
	//	error: 断开时获得的 lastError
	//	通过 lastError 可以判断网络断开的原因
	//
	// 用法:
	//	if (error == 0) {
	// 		if (disconnectedBySelf)
	// 			cout << "主动断开" << endl;
	// 		else
	// 			cout << "对方断开" << endl;
	// 	} else {
	// 		if (error == 501) {
	// 			if (getInt ("last-time", -2) < 2000)
	// 				cout << "超时断开, 可能是对方原因" << endl;
	// 			else
	// 				cout << "超时断开, 可能是己方原因" << endl;
	// 		}
	// 	}
	virtual void onDisconnected (int error) {
	}

	// 功能:
	//	玩家事件
	//
	// 参数:
	//	clientIndex: 玩家索引值
	//	e: 事件, 包括: 玩家新加入("new"), 离开("left"), 回来("return"), "lag"
	//
	// 用法:
	//	例如用于提示某个玩家进入或离开
	virtual void onPlayer (uint8_t clientIndex, string e) {
	}
};

#endif // NANOCLIENT_H_
