/*
功能:
	封装 Nanolink SDK 基础接口 并 增加事件接口

用法:
	#1, 从 NanoClient 派生 MyClient, 实现必要的 onMessage 事件接口
	var myClient = new NanoClient ({
		// 收到数据
		onMessage: function (data, fromIndex) {
			...
		},

		onResync (fromIndex) {
			...
		}
	});

	#2, 初始化
	myClient.init("***appKey***", 2);
	myClient.config("time-machine", "true");

	#3, 连接 (通常在"开始"按钮触发)
	myClient.connect(1);

	#4, 在 Update 中调用 doUpdate 接口接收数据处理状态
	myClient.doUpdate ();

	#5, 发送数据 (在事件产生的地方序列化数据并调用 send)
	myClient.send (byte* data, int len);

	#6, 断开连接 (通常在战斗结束的时候)
	myClient.disconnect ();
*/

function NanoClient(params) {
	var newClient = {
		status: "",
		disconnectedBySelf: false, // 是否主动断开
		players: 0, // 最大玩家索引
		last_times: new Array(64), // 玩家上次事件时间 (多少 ms 之前)

		// 功能:
		//	发送一个协议"分隔符"标记, 用在多人联网时
		//	** 已经在 onResync 前自动调用, 不需要再调用该接口
		//
		// 参数:
		//	mark: 标记值, 保留, 通常为0
		//
		// 返回值:
		// 	成功: >= 0, 发送数据的索引值
		//	失败: -1, 无法添加任务到发送队列满
		//		可以用 config ("send-tasks-size", "xxx") 调整发送队列大小
		_sendmark: function (mark) {
			return nano_sendmark(mark);
		},

		// 功能:
		//	SDK 初始化
		//
		// 参数:
		//	appKey: 应用ID
		//	mode: SDK 模式
		//
		//	mode 定义:
		// 		0: 回放
		// 		1: 单人, 仅用于记录
		// 		2: 1对1联网方式(缺省)
		// 		3: 多人联网方式
		//
		// 返回值:
		//	成功: 0
		//	失败: -1, mode 参数错误时失败
		init: function (appKey, mode) {
			this.disconnectedBySelf = false;

			this.players = 0;
			for (var i = 0; i < 64; i++)
				this.last_times[i] = -1;

			return nano_init(appKey, mode);
		},

		// 功能:
		//	设置 协议中的一些参数, 具体参看 "Nanolink SDK 参数设置" 文档
		//
		// 参数:
		//	name: 参数名称
		//	value: 参数值
		//
		// 返回值:
		//	成功: 0
		// 	失败: -1, 不支持的参数名称
		config: function (name, value) {
			return nano_config(name, value);
		},

		// 功能:
		//	外网匹配接口
		//
		// 参数:
		// 	v: 如果为数值, 客户端的等级, 具体匹配等级范围在服务器设置
		//	v: 如果为字符串, 匹配的房间名称, 如果为空，局域网匹配
		//	mode: 等级匹配可以分为不同模式, 支持 0-15 模式, 缺省为0
		//		例如, mode=0, 代表 经典模式, mode=1, 代表冒险模式, ...
		// 	timeout: 匹配等待时间, 单位秒，默认15秒
		//
		// 返回值:
		//	成功: 0
		//	失败: -1, init 接口没有调用 或 端口打开失败
		connect: function (v, mode, timeout) {
			mode = mode || 0;
			timeout = timeout || 15;
			if (typeof v == "string")
				return nano_connect2(v, timeout * 1000);
			else
				return nano_connect(v, mode, timeout * 1000);
		},

		// 功能:
		//	断开当前连接
		//
		// 返回值:
		//	成功: 0
		// 	失败: -1, init 接口没有调用
		disconnect: function () {
			this.disconnectedBySelf = true;
			return nano_disconnect();
		},

		// 功能:
		//	发送数据
		//
		// 参数:
		//	data: 待发送的数据
		//	size: 发送数据的长度
		//	strength: 发送广播强度, 仅在多人联网时有效
		//		0 发送给主机, 1-254 相同空间区域广播, 255 所有空间广播
		//
		// 返回值:
		// 	成功: >= 0, 发送数据的索引值
		//	失败: -1, 无法添加任务到发送队列满
		//		可以用 config ("send-tasks-size", "xxx") 调整发送队列大小
		send: function (data, strength) {
			strength = strength || 255;
            // data 参数不能直接使用 Uint8Array，需转换到数组中
			return nano_send(Array.from(data), strength);
		},

		// 功能:
		//	接收数据
		//	建议用 doUpdate + onMessage 方式接收处理数据
		//
		// 参数:
		//	data: 接收数据缓冲区
		//	size: 接收数据缓冲区大小
		// 	clientIndex: 发送者的索引值
		//
		// 返回值:
		// 	返回数据的长度
		//	-1: 没有收到新的数据
		//	0: 收到 resync 指令
		//	1: 收到 mark 指令
		//	>1: 收到用户指令
		recv: function () {
			var msg = nano_recv();
			if (!msg)
				return null;

            // data 数据将数组 转为 Uint8Array。 send() 反向
			return {
				data: new Uint8Array(msg[0]),
				fromIndex: msg[1]
			};
		},

		// 功能:
		//	更新客户端的位置(x,y), 空间(z), 接收增益(k)
		//
		// 参数:
		//	x, y 玩家的位置
		//	z 玩家所在空间, 缺省为0, 0-255, 不同空间互相不会收到数据
		//	k 接收增益, 接收范围的放大倍数
		//
		// 返回值:
		//	成功返回 0
		update: function(x, y, z, k) {
			z = z || 0;
			k = k || 1;
			return nano_update(x, y, z, k);
		},

		// 功能:
		//	加载日志数据, 用于日志回放
		//
		// 参数:
		//	filePath: 存放日志数据的文件路径
		//
		// 返回值:
		//	成功: 返回文件长度
		//	失败: -1, init 没有调用 或 文件打开失败
		load: function (filePath) {
			return nano_load(filePath, -1);
		},

		// 功能:
		//	保存日志数据
		//	需要先启用日志功能 config ("time-machine", "true");
		//
		// 参数:
		//	filePath: 存放日志数据的文件路径
		//
		// 返回值:
		//	成功: 返回文件长度
		//	失败: -1, init 没有调用 或 文件打开失败
		save: function (filePath) {
			return nano_save(filePath, -1);
		},

		// 功能:
		//	获取协议中的一些数值(整数类型)
		//
		// 参数:
		// 	name: players, latency, last-time, last-error, client-index, is-master, master-index, status
		//  n: last-time 时配合使用, 缺省-1
		//
		// 返回值:
		//	该参数的数值, 如果该参数不存在, 返回0
		getInt: function (name, n) {
			// 默认值 -1
			n = (typeof n !== 'undefined') ? n : -1;

			if (name == "players") {
				if (n == -1)
					return this.players;
				else {
					// 房间内在线玩家数量, "在线" 根据 n(timeout) 判断
					var count = 0;
					for (var i = 0; i < this.players; i ++) {
						var v = nano_getInt ("last-time", i);
						if (v >= 0 && v < n) {
							count ++;
						}
					}
	
					return count;
				}
			}
	
			return nano_getInt(name, n);
		},

		// 功能:
		//	获取协议中的一些数值(字符串类型)
		//
		// 参数:
		// 	name: server-id, target-id, client-id, status, stats
		//
		// 返回值:
		//	该参数的数值, 如果该参数不存在, 返回""
		getString: function (name) {
			return nano_getString(name);
		},

		// 功能:
		//	获得当前协议状态
		//
		// 返回值:
		//	init, matching, connecting, connected, disconnect
		getStatus: function () {
			return this.getString("status");
		},

		// 功能:
		//	判断连接是否建立
		//
		// 返回值:
		//	true: 已经建立连接
		//	false: 还没有建立连接
		isConnected: function () {
			return (this.getStatus() == "connected");
		},

		// 功能:
		//	判断连接是否断开
		//
		// 返回值:
		//	true: 已经断开连接
		//	false: 还没有断开连接
		isDisconnect: function () {
			return (this.getStatus() == "disconnect");
		},

		// 功能:
		//	处理状态变化与数据接收, 分发到接口的各个 on 事件中
		//
		// 返回值:
		//	>=0: 接收并分发的数据条数
		//	-1: 连接还没有建立
		doUpdate: function () {
			// 状态变化时, 调用相应状态变化函数
			if (this.status != this.getStatus()) {
				var newStatus = this.getStatus();
                console.log ("newStatus: " + newStatus + ", oldStatus: " + this.status);

				if (newStatus == "connected") {
					this.onConnected();
					
					// 匹配成功时，根据索引号计算当前房间人数
					this.players = this.getInt ("client-index") + 1;
					
                    this._sendmark(0);
					
                    this.onResync(this.getInt("client-index"));
				}

				if (newStatus == "disconnect") {
					this.onDisconnected(this.getInt("last-error"));
				}

				this.onStatusChanged(this.getStatus(), this.status);

				this.status = newStatus;
			}

			// 还没有连接成功，忽略后续的接收处理
			if (!this.isConnected())
				return -1;

			// 检查是否有人 "离开" 或 "回来"
			var timeout = 2000;
			for (var i = 0; i < this.players; i++) {
				var old_last_time = this.last_times[i];
				this.last_times[i] = this.getInt("last-time", i);

				if (old_last_time < timeout && this.last_times[i] > timeout) {
					this.onPlayer(i, "left"); // 离开
				} else if (old_last_time > timeout && this.last_times[i] < timeout) {
					this.onPlayer(i, "return"); // 回来
				}
			}

			var count = 0;
			while (true) {
				var msg = this.recv();
				if (msg == null)
					break;

				if (msg.fromIndex >= this.players) {
					// 有可能 msg.fromIndex 一下子比 this.players 大不少
					for (var i = this.players; i < msg.fromIndex + 1; i++) {
						this.last_times[i] = this.getInt("last-time", i);

						this.onPlayer(i, "new");
					}
					this.players = msg.fromIndex + 1;
				}

				if (msg.data.length == 0) { // 收到同步指令
					this._sendmark(0);
					this.onResync(msg.fromIndex);
				} else if (msg.data.length == 1) { // 收到标记指令
					// this.onMark (msg.data [0]);
				} else { // 收到数据
					this.onMessage(msg.data, msg.fromIndex);
				}

				count++;
			}

			return count;
		},

		// 功能:
		//	接收到一个消息时的处理接口
		//
		// 参数:
		//	data: 接收数据缓冲区
		//	size: 数据大小
		//	fromIndex: 发送者
		//
		// 用法:
		//	收到的数据进行解码, 转换成不同指令, 并分发给对应的指令代码
		onMessage: function (data, fromIndex) {},

		// 功能:
		//	收到 resync 指令时的处理接口
		//
		// 参数:
		//	fromIndex: 发送者
		//	如果是自己(建立连接时触发), fromIndex = getInt ("client-index")
		//
		// 用法:
		//  向其他客户端重新发送当前客户端的完整数据
		//	包括 昵称, 颜色, ... 等只在初始化传输的数据
		onResync: function (fromIndex) {},

		// 功能:
		//	状态变化
		//
		// 参数:
		//	newStatus: 变化后的状态
		//	oldStatus: 变化前的状态
		onStatusChanged: function (newStatus, oldStatus) {},

		// 功能:
		//	连接建立
		//
		// 用法:
		//	连接建立后初始化, 开始游戏
		onConnected: function () {},

		// 功能:
		//	连接断开
		//
		// 参数:
		//	error: 断开时获得的 lastError
		//	通过 lastError 可以判断网络断开的原因
		//
		// 用法:
		//	if (error == 0) {
		// 		if (disconnectedBySelf)
		// 			cout << "主动断开" << endl;
		// 		else
		// 			cout << "对方断开" << endl;
		// 	} else {
		// 		if (error == 501) {
		// 			if (getInt ("last-time", -2) < 2000)
		// 				cout << "超时断开, 可能是对方原因" << endl;
		// 			else
		// 				cout << "超时断开, 可能是己方原因" << endl;
		// 		}
		// 	}
		onDisconnected: function (error) {},

		// 功能:
		//	玩家事件
		//
		// 参数:
		//	clientIndex: 玩家索引值
		//	e: 事件, 包括: 玩家新加入("new"), 离开("left"), 回来("return"), "lag"
		//
		// 用法:
		//	例如用于提示某个玩家进入或离开
		onPlayer: function (clientIndex, e) {
		}
	}

	if (typeof params == 'object') {
		for (var attrname in params) {
			newClient[attrname] = params[attrname];
		}
	}

	return newClient;
}
