
#ifndef NANOLINK_H_
#define NANOLINK_H_

#include <stdint.h>
#include <stddef.h>

#define STATUS_DISCONNECT	0
#define STATUS_CONNECTING	1
#define STATUS_CONNECTED	2
#define STATUS_MATCHING		3
#define STATUS_INIT			9

#ifdef __cplusplus
extern "C" {
#endif
	// 功能:
	//	SDK 初始化
	//
	// 参数:
	//	appKey: 应用ID
	//	mode: SDK 模式
	//
	//	mode 定义:
	// 		0: 回放
	// 		1: 单人, 仅用于记录
	// 		2: 1对1联网方式(缺省)
	// 		3: 多人联网方式
	//
	// 返回值:
	//	成功: 0
	//	失败: -1, mode 参数错误时失败
	int32_t nano_init (const char* appKey, int mode = 2);

	// 功能:
	//	设置 协议中的一些参数, 具体参看 "Nanolink SDK 参数设置" 文档
	//
	// 参数:
	//	name: 参数名称
	//	value: 参数值
	//
	// 返回值:
	//	成功: 0
	// 	失败: -1, 不支持的参数名称
	int32_t nano_config (const char* name, const char* value);

	// 功能:
	//	基于等级匹配连接(外网)
	//
	// 参数:
	// 	level: 客户端的等级, 具体匹配等级范围在服务器设置
	//	mode: mode 低4位(0-15) 用于表示等级匹配的不同模式, 支持 0-15 模式, 缺省为0. 例如, mode=0, 代表 经典模式, mode=1, 代表冒险模式, ..
	//		例如, mode=0, 代表 经典模式, mode=1, 代表冒险模式, ...
	//		mode 高4位用来表示加入/创建房间
	//		mode >> 4 == 0: 加入或创建房间
	//		mode >> 4 == 1: 加入房间
	//		mode >> 4 == 2: 创建房间
	// 	timeout: 匹配等待时间, 单位秒，默认15秒
	//
	// 返回值:
	//	成功: 0
	//	失败: -1, init 接口没有调用 或 端口打开失败
	int32_t nano_connect (uint8_t level, uint8_t mode = 0, int32_t timeout=15000);

	// 功能:
	//	基于名称匹配连接
	//
	// 参数:
	//	group: 匹配的房间名称
	//		group 为空，通过局域网广播方式连接
	//		group 不为空，通过服务器建立相同 group 之间用户的连接
	//		group 名称可以重用，如果group在服务器不存在，创建并等待其它玩家连接
	//		group 首字母为 ‘+’，表示被邀请加入, 此时如果 group 在服务器不存在，立即返回连接失败，错误代码 404
	//		group 首字母为 ‘^’，表示创建房间, 此时如果 group 在服务器已经存在，立即返回连接失败，错误代码 404
	// 	timeout: 匹配等待时间, 单位秒，默认15秒
	//
	// 返回值:
	//	成功: 0
	//	失败: -1, init 接口没有调用 或 端口打开失败
	int32_t nano_connect2 (const char* group = NULL, int32_t timeout=15000);

	// 功能:
	//	断开当前连接
	//
	// 返回值:
	//	成功: 0
	// 	失败: -1, init 接口没有调用
	int32_t nano_disconnect ();

	// 功能:
	//	发送数据
	//
	// 参数:
	//	data: 待发送的数据
	//	size: 发送数据的长度
	//	strength: 发送广播强度, 仅在多人联网时有效
	//		0 发送给主机, 1-254 相同空间区域广播, 255 所有空间广播
	//
	// 返回值:
	// 	成功: >= 0, 发送数据的索引值
	//	失败: -1, 无法添加任务到发送队列满
	//		可以用 config ("send-tasks-size", "xxx") 调整发送队列大小
	int32_t nano_send (uint8_t* data, uint32_t size, uint8_t strength = 255);

	// 功能:
	//	发送一个协议"分隔符"标记, 用在多人联网时
	//
	// 参数:
	//	mark: 标记值, 保留, 通常为0
	//
	// 返回值:
	// 	成功: >= 0, 发送数据的索引值
	//	失败: -1, 无法添加任务到发送队列满
	//		可以用 config ("send-tasks-size", "xxx") 调整发送队列大小
	int32_t nano_sendmark (uint8_t mark = 0);

	// 功能:
	//	接收数据
	//	建议用 doUpdate + onMessage 方式接收处理数据
	//
	// 参数:
	//	data: 接收数据缓冲区
	//	size: 接收数据缓冲区大小
	// 	clientIndex: 发送者的索引值
	//
	// 返回值:
	// 	返回数据的长度
	//	-1: 没有收到新的数据
	//	0: 收到 resync 指令
	//	1: 收到 mark 指令
	//	>1: 收到用户指令
	int32_t nano_recv (uint8_t* data, uint32_t size, uint8_t& clientIndex);

	// 功能:
	//	更新客户端的位置(x,y), 空间(z), 接收增益(k)
	//
	// 参数:
	//	x, y 玩家的位置
	//	z 玩家所在空间, 缺省为0, 0-255, 不同空间互相不会收到数据
	//	k 接收增益, 接收范围的放大倍数
	//
	// 返回值:
	//	成功返回 0
	int32_t nano_update (uint8_t x, uint8_t y, uint8_t z = 0, uint8_t k = 1);

	// 功能:
	//	获取协议中的一些数值(整数类型)
	//
	// 参数:
	// 	name: players, latency, last-time, last-error, client-index, is-master, master-index, status
	//  n: last-time 时配合使用, 缺省-1
	//
	// 返回值:
	//	该参数的数值, 如果该参数不存在, 返回0
	int32_t nano_getInt (const char* name, int n = -1);

	// 功能:
	//	获取协议中的一些数值(字符串类型)
	//
	// 参数:
	// 	name: server-id, target-id, client-id, status, stats
	//
	// 返回值:
	//	该参数的数值, 如果该参数不存在, 返回""
	int32_t nano_getString (const char* name, char* data, uint32_t size);

	// 功能:
	//	加载日志数据, 用于日志回放
	//
	// 参数:
	//	data: 存放日志数据的缓冲区, 如果 size = -1, data 作为文件路径
	//	size: 缓冲区大小
	//
	// 返回值:
	//	成功: 返回数据大小
	//	失败: -1, init 没有调用 或 文件打开失败
	int32_t nano_load (uint8_t* data, uint32_t size = -1);

	// 功能:
	//	保存日志数据
	//	需要先启用日志功能 config ("time-machine", "true");
	//
	// 参数:
	//	data: 存放日志数据的缓冲区, 如果 size = -1, data 作为文件路径
	//	size: 缓冲区大小
	//
	// 返回值:
	//	成功: 返回数据大小
	//	失败: -1, init 没有调用 或 文件打开失败
	int32_t nano_save (uint8_t* data, uint32_t size = -1);
#ifdef __cplusplus
}
#endif

#endif // NANOLINK_H_
