require "nanolink"

_G.NanoClient = {
    status = "",
    disconnectedBySelf = false, --是否主动断开
    players = 0,    -- 最大玩家索引
    last_times = {},  -- 玩家上次事件时间(多少 ms 之前)
}

--[[ 
init
功能:
    SDK 初始化
参数:
    appKey: 应用ID
    mode: SDK 模式
mode 定义:
        0: 回放
        1: 单人, 仅用于记录
        2: 1对1联网方式(缺省)
        3: 多人联网方式
返回值:
    成功: 0
    失败: -1, mode 参数错误时失败
]]--
NanoClient.init = function (appKey, mode)
    NanoClient.disconnectedBySelf = false

    NanoClient.players = 0
    for i=1, 64 do
        NanoClient.last_times[i] = -1
    end

    return nanolink.init(appKey, mode)
end

--[[
config
    功能:
    设置 协议中的一些参数, 具体参看 "Nanolink SDK 参数设置" 文档

    参数:
    name: 参数名称
    value: 参数值

    返回值:
    成功: 0
    失败: -1, 不支持的参数名称
]]--
NanoClient.config = function (name, value)
    return nanolink.config(name, value)
end

--[[
    功能:
    外网匹配接口

    参数:
    v: 如果为数值, 客户端的等级, 具体匹配等级范围在服务器设置
    v: 如果为字符串, 匹配的房间名称, 如果为空，局域网匹配
    mode: 等级匹配可以分为不同模式, 支持 0-15 模式, 缺省为0
        例如, mode=0, 代表 经典模式, mode=1, 代表冒险模式, ...
    timeout: 匹配等待时间, 单位秒，默认15秒

    返回值:
    成功: 0
    失败: -1, init 接口没有调用 或 端口打开失败
]]--    
NanoClient.connect = function (v, mode, timeout)
    mode = mode or 0
    timeout = timeout or 15
    if type(v) == "string" then
        return nanolink.connect2(v, timeout * 1000)
    else
        return nanolink.connect(v, mode, timeout * 1000)
    end
end

--[[
    功能:
    断开当前连接

    返回值:
    成功: 0
    失败: -1, init 接口没有调用
]]
NanoClient.disconnect = function ()
    NanoClient.disconnectedBySelf = true
    return nanolink.disconnect()    
end

--[[
    功能:
    发送数据
    
    参数:
    data: 待发送的数据
    size: 发送数据的长度
    strength: 发送广播强度, 仅在多人联网时有效
         0 发送给主机, 1-254 相同空间区域广播, 255 所有空间广播
    
    返回值:
    成功: >= 0, 发送数据的索引值
    失败: -1, 无法添加任务到发送队列满
        可以用 config ("send-tasks-size", "xxx") 调整发送队列大小
]]
NanoClient.send = function (data, strengh)
    strengh = strengh or 255
    return nanolink.send(data, strengh)
end

-- 功能:
--	发送一个协议"分隔符"标记, 用在多人联网时
--	** 已经在 onResync 前自动调用, 不需要再调用该接口

-- 参数:
--	mark: 标记值, 保留, 通常为0
-- 返回值:
-- 	成功: >= 0, 发送数据的索引值
--	失败: -1, 无法添加任务到发送队列满
--		可以用 config ("send-tasks-size", "xxx") 调整发送队列大小
NanoClient.sendmark = function (mark)
    return nanolink.sendmark(mark)
end

--[[
    功能:
    接收数据
    建议用 doUpdate + onMessage 方式接收处理数据

    参数:
    data: 接收数据缓冲区
    size: 接收数据缓冲区大小
    clientIndex: 发送者的索引值

    返回值:
    返回数据的长度
    -1: 没有收到新的数据
    0: 收到 resync 指令
    1: 收到 mark 指令
    >1: 收到用户指令
]]
NanoClient.recv = function ()
    local recv_data, playerIndex = nanolink.recv()
    if recv_data == nil then
        return nil
    end

    local msg = {
        data = recv_data,
        fromIndex = playerIndex
    }

    return msg
end

--[[
    功能:
    更新客户端的位置(x,y), 空间(z), 接收增益(k)
    
    参数:
    x, y 玩家的位置
    z 玩家所在空间, 缺省为0, 0-255, 不同空间互相不会收到数据
    k 接收增益, 接收范围的放大倍数
    
    返回值:
    成功返回 0
]]--
NanoClient.update = function (x, y, z, k)
    local newz = z or 0
    local newk = k or 1
    return nanolink.update(x, y, newz, newk)
end

--[[
    功能:
    加载日志数据, 用于日志回放

    参数:
    filePath: 存放日志数据的文件路径

    返回值:
    成功: 返回文件长度
    失败: -1, init 没有调用 或 文件打开失败
]]--
NanoClient.load = function (filePath)
    return nanolink.load(filePath, -1)
end

--[[
    功能:
    保存日志数据
    需要先启用日志功能 config ("time-machine", "true");

    参数:
    filePath: 存放日志数据的文件路径

    返回值:
    成功: 返回文件长度
    失败: -1, init 没有调用 或 文件打开失败
]]
NanoClient.save = function (filePath)
    return nanolink.save(filePath, -1)
end

--[[
    功能:
    获取协议中的一些数值(整数类型)

    参数:
    name: players, latency, last-time, last-error, client-index, is-master, master-index, status
    n: last-time 时配合使用, 缺省-1

    返回值:
    该参数的数值, 如果该参数不存在, 返回0
]]
NanoClient.getInt = function (name, n)
    -- 默认值 -1
    if n == nil then n = -1 end

    if name == "players" then
        if n == -1 then
            return NanoClient.players
        else
            local count = 0;
            for i,v in ipairs(NanoClient.players) do
                local v = nanolink.getInt("last-time", i)
                if v >= 0 and v < n then
                    count = count + 1
                end
            end
            return count
        end
    end

    return nanolink.getInt(name, n)
end

--[[
    功能:
    获取协议中的一些数值(字符串类型)

    参数:
    name: server-id, target-id, client-id, status, stats

    返回值:
    该参数的数值, 如果该参数不存在, 返回""
]]
NanoClient.getString = function (name)
    return nanolink.getString(name)
end

--[[
    功能:
    获得当前协议状态

    返回值:
    init, matching, connecting, connected, disconnect
]]
NanoClient.getStatus = function ()
    return nanolink.getString("status")
end

--[[
    功能:
    判断连接是否建立

    返回值:
    true: 已经建立连接
    false: 还没有建立连接
]]
NanoClient.isConnected = function ()
    return NanoClient.getStatus() == "connected"
end

--[[
    功能:
    判断连接是否断开

    返回值:
    true: 已经断开连接
    false: 还没有断开连接
]]
NanoClient.isDisconnect = function ()
    return NanoClient.getStatus() == "disconnect"
end

-------------------------需要自定义的------------------------------------------
--[[
功能:
    连接断开

    参数:
    error: 断开时获得的 lastError
    通过 lastError 可以判断网络断开的原因
]]--
NanoClient.onStatusChanged = function (newStatus, oldStatus) 
end

--[[
    功能:
    连接建立

    用法:
    连接建立后初始化, 开始游戏
]]--
NanoClient.onConnected = function ()
end

--[[
    功能:
    连接断开

    参数:
    error: 断开时获得的 lastError
    通过 lastError 可以判断网络断开的原因
]]--
NanoClient.onDisconnected = function (reason)
end

--[[
    功能:
    接收到一个消息时的处理接口

    参数:
    data: 接收数据缓冲区
    size: 数据大小
    fromIndex: 发送者

    用法:
    收到的数据进行解码, 转换成不同指令, 并分发给对应的指令代码
]]--
NanoClient.onMessage = function (data, fromIndex)
end

--[[
    功能:
    收到 resync 指令时的处理接口

    参数:
    fromIndex: 发送者
    如果是自己(建立连接时触发), fromIndex = getInt ("client-index")

    用法:
    向其他客户端重新发送当前客户端的完整数据
    包括 昵称, 颜色, ... 等只在初始化传输的数据
]]
NanoClient.onResync = function (fromIndex)
end

--[[
    功能:
    玩家事件

    参数:
    clientIndex: 玩家索引值
    e: 事件, 包括: 玩家新加入("new"), 离开("left"), 回来("return"), "lag"

    用法:
    例如用于提示某个玩家进入或离开
]]
NanoClient.onPlayer = function (clientIndex, e)
end
-------------------------需要自定义的------------------------------------------

--[[
    功能:
    处理状态变化与数据接收, 分发到接口的各个 on 事件中

    返回值:
    >=0: 接收并分发的数据条数
    -1: 连接还没有建立
]]--
NanoClient.doUpdate = function ()
        -- 状态变化时, 调用相应状态变化函数
    if NanoClient.status ~= NanoClient.getStatus() then
        local newStatus = NanoClient.getStatus()
        print("newStatus:"..newStatus..",oldStatus:"..NanoClient.status)

        if newStatus == "connected" then
            NanoClient.onConnected()

            NanoClient.sendmark(0)

            NanoClient.onResync(NanoClient.getInt("client-index"))
        end

        if newStatus == "disconnect" then
            NanoClient.onDisconnected(NanoClient.getInt("last_error"))
        end

        NanoClient.onStatusChanged(NanoClient.getStatus(), NanoClient.status)
        NanoClient.status = newStatus
    end

    -- 还没有连接成功, 忽略后续的接受处理
    if not NanoClient.isConnected() then
        return -1
    end

    -- 检查是否有人 "离开" 或 "回来"
    local timeout = 2000
    for i=1, NanoClient.players do
        local old_last_time = NanoClient.last_times[i]
        NanoClient.last_times[i] = NanoClient.getInt("last-time", i)

        if old_last_time < timeout and NanoClient.last_times[i] > timeout then
            NanoClient.onPlayer(i, "left")   -- 离开
        elseif old_last_time > timeout and NanoClient.last_times[i] < timeout then
            NanoClient.onPlayer(i, "return")  -- 回来
        end
    end

    local count = 0
    while true do
        local msg = NanoClient.recv()
        if msg == nil then
            break
        end

        if msg.fromIndex >= NanoClient.players then
            -- 有可能 msg.fromIndex 一下子比 this.players 大不少
            for i=NanoClient.players, msg.fromIndex do
                NanoClient.last_times[i] = NanoClient.getInt("last-time", i)
                NanoClient.onPlayer(i, "new")
            end
            NanoClient.players = msg.fromIndex + 1
        end

        if string.len(msg.data) == 0 then     -- 收到指令
            NanoClient.sendmark(0)
            NanoClient.onResync(msg.fromIndex)
        elseif string.len(msg.data) == 1 then  -- 收到标记指令

        else
            NanoClient.onMessage(msg.data, msg.fromIndex)
        end

        count = count + 1
    end

    return count
end

NanoClient.extend = function (params)
    if type(params) == "table" then
        for k,v in pairs(params) do
            if type(v) == "function" then
                NanoClient[k] = v
            end
        end
    end

    return NanoClient
end
